package meerkat;

import com.google.protobuf.Timestamp;
import meerkat.bulletinboard.AsyncBulletinBoardClient;
import meerkat.crypto.DigitalSignature;
import meerkat.protobuf.BulletinBoardAPI.BulletinBoardMessage;
import meerkat.protobuf.BulletinBoardAPI.UnsignedBulletinBoardMessage;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.VoterRegistry.GroupID;
import meerkat.protobuf.VoterRegistry.VoterID;
import meerkat.protobuf.VoterRegistry.VoterInfo;
import meerkat.protobuf.VoterRegistry.VoterRegistryMessage;
import meerkat.registry.LatestMessagesCallBack;
import meerkat.registry.MessageCollectionUtils;
import meerkat.registry.MessagesCallBack;
import meerkat.registry.RegistryTags;

import java.io.InputStream;
import java.security.SignatureException;
import java.security.cert.CertificateException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * TODO : add ability to use DB of certificates
 */

/**
 * Created by Vladimir Eliezer Tokarev on 1/8/2016.
 * Gives the ability to synchronously manage voters information
 */
public class Registry implements VoterRegistry{

    protected Collection<DigitalSignature> signers;
    protected AsyncBulletinBoardClient bulletinBoardClient ;
    protected InputStream certificateStream;

    public void init(Collection<DigitalSignature> signers,
                     AsyncBulletinBoardClient communicator,
                     InputStream certificateStream) {
        this.signers = signers;
        this.bulletinBoardClient  = communicator;
        this.certificateStream = certificateStream;
    }

    /**
     * Creates BulletinBoardMessage with signed basicMessage and
     *         UnsignedBulletinBoardMessage that contains the basic message
     *
     * @param basicMessage BasicMessage
     * @return BulletinBoardMessage
     */
    protected BulletinBoardMessage createBulletinBoardMessage(UnsignedBulletinBoardMessage basicMessage) {
        try {
            BulletinBoardMessage.Builder bulletinBoardMessage = BulletinBoardMessage.newBuilder();
            for (DigitalSignature signer : signers) {
                signer.updateContent(basicMessage);
                Crypto.Signature signature = signer.sign();
                bulletinBoardMessage.setMsg(basicMessage).addSig(signature);

                signer.loadVerificationCertificates(certificateStream);
            }
            return bulletinBoardMessage.build();
        } catch (SignatureException | CertificateException e) {
            e.printStackTrace();
            return null;
        }
    }

    public void addVoter(VoterInfo voterInfo, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder().
                        addTag(RegistryTags.ID_TAG + voterInfo.getId().getId())
                        .addTag(RegistryTags.VOTER_ENTRY_TAG)
                        .addTag(RegistryTags.VOTER_DATA_TAG + voterInfo.getInfo())
                        .setTimestamp(Timestamp.newBuilder().setNanos((int) System.nanoTime()).build());

        bulletinBoardClient.postMessage(createBulletinBoardMessage(basicMessage.build()), new MessagesCallBack(callback));
    }

    public void addToGroups(VoterRegistryMessage voterGroup, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterGroup.getVoterID().getId())
                        .addTag(RegistryTags.ADD_TO_GROUP_TAG)
                        .setTimestamp(Timestamp.newBuilder().setNanos((int) System.nanoTime()).build());

        for (GroupID groupId : voterGroup.getGroupIDList())
        {
            basicMessage.addTag(RegistryTags.GROUP_ID_TAG + groupId.getId());
        }

        bulletinBoardClient.postMessage(createBulletinBoardMessage(basicMessage.build()), new MessagesCallBack(callback));
    }

    public void setVoted(VoterID voterId, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterId.getId())
                        .addTag(RegistryTags.VOTE_ACTION_TAG)
                        .setTimestamp(Timestamp.newBuilder().setNanos((int) System.nanoTime()).build());

        bulletinBoardClient.postMessage(createBulletinBoardMessage(basicMessage.build()), new MessagesCallBack(callback));
    }

    public void getGroups(GroupID groupID, RegistryCallBack callback) {
        List<String> GroupsActionsTags = new ArrayList<String>(2) {{
           add(RegistryTags.GROUP_ID_TAG + groupID.getId());
        }};
        bulletinBoardClient.readMessages(MessageCollectionUtils.generateFiltersFromTags(GroupsActionsTags),
                new LatestMessagesCallBack(callback, signers));
    }

    public void getPersonIDDetails(VoterID voterID, RegistryCallBack callback) {
        List<String> GroupsActionsTags = new ArrayList<String>() {{
            add(RegistryTags.ID_TAG + voterID.getId());
            add(RegistryTags.VOTER_ENTRY_TAG);
        }};
        bulletinBoardClient.readMessages(MessageCollectionUtils.generateFiltersFromTags(GroupsActionsTags),
                new LatestMessagesCallBack(callback, signers));
    }
}
